#ifndef __ASM_HARDIRQ_H
#define __ASM_HARDIRQ_H

/* XXX much of this is wrong, esp for SMP The irq stuff needs to move to the psa */

#include <asm/psa.h>

/*
 * Are we in an interrupt context? Either doing bottom half
 * or hardware interrupt processing?
 */
#define in_interrupt()  (_PSA_.local_irq_count + _PSA_.local_bh_count != 0)

#ifndef __SMP__

#define i370_get_psa(cpu)  ((psa_t *) 0x0)

#define hardirq_trylock(cpu)	(i370_get_psa(cpu)->local_irq_count == 0)
#define hardirq_endlock(cpu)	do { } while (0)

#define hardirq_enter(cpu)	(i370_get_psa(cpu)->local_irq_count++)
#define hardirq_exit(cpu)	(i370_get_psa(cpu)->local_irq_count--)

#define synchronize_irq()	do { } while (0)

#else /* __SMP__ */

#include <asm/atomic.h>

/* implements routine to get the prefix page */
/* the prfix page for processor N will map to page N+1
 * i.e. cpu zero pfx page is on page 1.
 */
#define i370_get_psa(cpu)  i370_get_pas_is_not_implemented;

extern unsigned char global_irq_holder;
extern unsigned volatile int global_irq_lock;
extern atomic_t global_irq_count;

static inline void release_irqlock(int cpu)
{
	/* if we didn't own the irq lock, just ignore.. */
	if (global_irq_holder == (unsigned char) cpu) {
		global_irq_holder = NO_PROC_ID;
		clear_bit(0,&global_irq_lock);
	}
}

static inline void hardirq_enter(int cpu)
{
	psa_t *psa = i370_get_psa(cpu);
	psa->local_irq_count ++;
	atomic_inc(&global_irq_count);
}

static inline void hardirq_exit(int cpu)
{
	psa_t *psa = i370_get_psa(cpu);
	atomic_dec(&global_irq_count);
	psa->local_irq_count --;
}

static inline int hardirq_trylock(int cpu)
{
	return !atomic_read(&global_irq_count) && !test_bit(0,&global_irq_lock);
}

#define hardirq_endlock(cpu)	do { } while (0)

extern void synchronize_irq(void);

#endif /* __SMP__ */

#endif /* __ASM_HARDIRQ_H */
