#ifndef _I370_PGALLOC_H
#define _I370_PGALLOC_H
/* 
 * Page Table allocation stuff for the ESA/390.  
 * XXX under constructions, some things need fixin
 */

#include <linux/config.h>
#include <linux/mm.h>


extern __inline__ pgd_t *i370_pgd_alloc(void)
{
	pgd_t *ret;

	/* grab two pages, for 2048 entires mapping 2GB */
	ret = (pgd_t *)__get_free_pages(GFP_KERNEL,1);
	if (ret) {
		/* zero out the entire pgd, not just the 'user' part of it */
		/* memset (ret, 0, USER_PTRS_PER_PGD * sizeof(pgd_t)); */
		/* memset won't work, we need to set the invalid bit */
		int i;
		pmd_t *pme = (pmd_t *) ret;
		/* printk ("in pgd_alloc alloced 2 pages at %p\n", ret); */
		for (i=0; i<USER_PTRS_PER_PGD; i++) {
			pmd_clear (pme);
			pme ++;
		}
		
	}
	return ret;
}

extern __inline__ void i370_pgd_free(pgd_t *pgd)
{
	/* free two pages of the pgd */
	free_pages((unsigned long)pgd,1);
}


extern __inline__ void i370_pte_free(pte_t *pte)
{
	free_page((unsigned long)pte);
}

extern void __bad_pte(pmd_t *pmd);

#define pte_alloc(pmd,addr)     i370_pte_alloc(pmd,addr)
#define pte_free(pte)           i370_pte_free(pte)

#define pte_alloc_kernel	i370_pte_alloc
#define pte_free_kernel(pte)    i370_pte_free(pte)

#define pmd_free		i370_pmd_free
#define pmd_alloc		i370_pmd_alloc
#define pmd_free_kernel		i370_pmd_free
#define pmd_alloc_kernel	i370_pmd_alloc

#define pgd_alloc()             i370_pgd_alloc()
#define pgd_free(pgd)           i370_pgd_free(pgd)

extern pte_t *i370_get_pte(pmd_t *pmd, unsigned long pt_entry_index);

extern inline pte_t * i370_pte_alloc(pmd_t * pmd, unsigned long address)
{
	unsigned long entry_idx = (address >> PAGE_SHIFT) & (PTRS_PER_PTE - 1);
	if (pmd_none(*pmd)) {
		return i370_get_pte(pmd, entry_idx);
	}
	if (pmd_bad(*pmd)) {
		__bad_pte(pmd);
		return NULL;
	}
	return ((pte_t *)pmd_page(*pmd)) + entry_idx;
}

/*
 * allocating and freeing a pmd is trivial: the 1-entry pmd is
 * inside the pgd, so has no extra memory associated with it.
 */
extern inline void i370_pmd_free(pmd_t * pmd)
{
}

extern inline pmd_t * i370_pmd_alloc(pgd_t * pgd, unsigned long address)
{
	return (pmd_t *) pgd;
}


extern int do_check_pgt_cache(int, int);

/* XXX ??? what is this supposed to do ??? */
/* I think this is used only by vmalloc, and is used to make
 * sure that all kernel processes see the virtual page mapping.
 * right now we don't support vmalloc, for the moment.
 */
extern inline void set_pgdir(unsigned long address, pgd_t entry)
{
	printk ("i370_set_pgdir(): unimplemented\n");
	i370_halt();
#if 0
	struct task_struct * p;
        
	read_lock(&tasklist_lock);
	for_each_task(p) {
		if (!p->mm)
			continue;
		*pgd_offset(p->mm,address) = entry;
	}
	read_unlock(&tasklist_lock);
#endif 
}

#endif /* _I370_PGALLOC_H */
