#include <linux/config.h>
/* 
 * Page Table mamangement stuff for the ESA/390.  
 * XXX under constructions, some things need fixin
 */

#ifndef _I370_PGTABLE_H
#define _I370_PGTABLE_H

#ifndef __ASSEMBLY__
#include <asm/processor.h>		/* For TASK_SIZE */
#include <asm/mmu.h>
#include <asm/page.h>

extern void i370_flush_tlb_all(void);
extern void i370_flush_tlb_mm(struct mm_struct *mm);
extern void i370_flush_tlb_page(struct vm_area_struct *vma, unsigned long vmaddr);
extern void i370_flush_tlb_range(struct mm_struct *mm, unsigned long start,
			    unsigned long end);

#define flush_tlb_all   i370_flush_tlb_all
#define flush_tlb_mm    i370_flush_tlb_mm
#define flush_tlb_page  i370_flush_tlb_page
#define flush_tlb_range i370_flush_tlb_range

/*
 * No cache flushing is required on i370 architecture
 */
#define flush_cache_all()		do { } while (0)
#define flush_cache_mm(mm)		do { } while (0)
#define flush_cache_range(mm, a, b)	do { } while (0)
#define flush_cache_page(vma, p)	do { } while (0)

#define flush_icache_range(a,b)		do { } while (0)
#define flush_page_to_ram(a)		do { } while (0)

extern unsigned long va_to_phys(unsigned long address);
extern pte_t *va_to_pte(struct task_struct *tsk, unsigned long address);
#endif /* __ASSEMBLY__ */

/*
 * The ESA/390 MMU uses a page table containing PTEs, together with
 * a segment table containing STE's (pointers to page tables) to define
 * the virtual to physical address mapping. We collapse the linux pgd 
 * and pmd to be the same thing, and map them to the 390 segment table.
 *
 * Implementation note: Although conceptually, the pgd and pmd are collapsed
 * into one, the actual implementation is oddly split.  This is because many
 * important functions (such as i370_pte_alloc) get only the pmd as an argument, 
 * and must still 'do the right thing'.
 */

/* PMD_SHIFT determines the size of the area mapped by the second-level
 * page tables.  For the i370, it is a 1MB segment 
 */
#define PMD_SHIFT	20
#define PMD_SIZE	(1UL << PMD_SHIFT)
#define PMD_MASK	(~(PMD_SIZE-1))

/* PGDIR_SHIFT determines what a third-level page table entry can map */
#define PGDIR_SHIFT	20
#define PGDIR_SIZE	(1UL << PGDIR_SHIFT)
#define PGDIR_MASK	(~(PGDIR_SIZE-1))

/*
 * entries per page directory level: our page-table tree is two-level, so
 * we don't really have any PMD directory.
 */
#define PTRS_PER_PTE	256
#define PTRS_PER_PMD	1
#define PTRS_PER_PGD	2048

/* Because each i370 page table has only 256 entries, we can in fact
 * fit four page tables onto one page. (256*4 *4bytes per entry = 4096 bytes) */
#define PT_PER_PAGE	4

/* segment table entry page table origin mask */
#define SEG_PTO_MASK	0x7fffffc0

/* page frame real address mask */
#define PFRA_MASK	0x7ffff000

/* USER_PTRS_PER_PGD should equal 2048 to address 2GB */
#define USER_PTRS_PER_PGD	(TASK_SIZE / PGDIR_SIZE)

/* VMALLOC_OFFSET is an (arbitrary) offset to the start of the vmalloc 
 * kernel VM area: the current 1GB value means that all kernel vm addresses
 * will have a "1" in bit position 30.  
 * We do this to catch out-of-bounds real memory accesses.
 * It also means that we have imposed a practical limit of max of 1GB 
 * a real memory.  This limit can be raised, although it will decrease 
 * the amount of virtual memory available to the kernel.
 * 
 * The vmalloc() routines leaves a hole of 4kB between each vmalloced
 * area for the same reason. ;)
 *
 */
#define VMALLOC_OFFSET	(0x40000000) /* 1GB */
#define VMALLOC_START ((VMALLOC_OFFSET & ~(VMALLOC_OFFSET-1)))
#define VMALLOC_VMADDR(x) ((unsigned long)(x))
#define VMALLOC_END	0x7fffe000

/*
 * Bits in a linux-style PTE.  Two bits match the hardware
 * PTE, the other bits are software-only thingies that inhabit 
 * low byte of the pte and hopefull won't hurt anybody.
 */
#define _PAGE_INVALID	0x400	/* hardware pte: I bit is set */
#define _PAGE_RO	0x200	/* hardware pte: P bit (write protect) */

/* hardware segment table entry */
#define _SEG_INVALID    0x20	/* segment invalid bit */
#define _SEG_COMMON     0x10	/* segment common bit */

#define _PAGE_DIRTY	0x002	/* storage key C: page changed */
#define _PAGE_ACCESSED	0x004	/* storage key R: page referenced */

#define _PAGE_USER	0x001	/* software: user-mode */
#define _PAGE_SHARED	0x010

#define _PAGE_CHG_MASK	(PAGE_MASK | _PAGE_ACCESSED | _PAGE_DIRTY)

#define PAGE_NONE	__pgprot( _PAGE_INVALID)

#define PAGE_SHARED	__pgprot( _PAGE_USER | _PAGE_SHARED)
#define PAGE_COPY	__pgprot( _PAGE_USER)
#define PAGE_READONLY	__pgprot( _PAGE_USER | _PAGE_RO)
#define PAGE_KERNEL	__pgprot( _PAGE_SHARED)
#define PAGE_KERNEL_CI	__pgprot( _PAGE_SHARED )

/*
 * We consider execute permission the same as read.
 * Also, write permissions imply read permissions.
 * If we go fancy, maybe we can do better but this will hold for now.
 * P macros imply copy-on-write, S macros for shared access
 *
 * XXX Some future data, we should consider using sske to 
 * mark pages exectable.  Thus data segments would get loaded
 * with a data storage key, text segments with a different key (?)
 * This would give nice security protection against buffer overrun 
 * stack-smashing attacks.
 */
#define __P000	PAGE_NONE
#define __P001	PAGE_READONLY
#define __P010	PAGE_COPY
#define __P011	PAGE_COPY
#define __P100	PAGE_READONLY
#define __P101	PAGE_READONLY
#define __P110	PAGE_COPY
#define __P111	PAGE_COPY

#define __S000	PAGE_NONE
#define __S001	PAGE_READONLY
#define __S010	PAGE_SHARED
#define __S011	PAGE_SHARED
#define __S100	PAGE_READONLY
#define __S101	PAGE_READONLY
#define __S110	PAGE_SHARED
#define __S111	PAGE_SHARED

/*
 * BAD_PAGETABLE is used when we need a bogus page-table, while
 * BAD_PAGE is used for a bogus page.
 */
#ifndef __ASSEMBLY__

extern pte_t __bad_page(void);
extern pte_t * __bad_pagetable(void);

#define BAD_PAGETABLE	__bad_pagetable()
#define BAD_PAGE	__bad_page()

/*
 * ZERO_PAGE is a global shared page that is always zero: used
 * for zero-mapped memory areas etc.  Note that this macor returns
 * a pointer to mem_map_t in the kernel page map array.
 */
extern unsigned long empty_zero_page[1024];
#define ZERO_PAGE(vaddr) (mem_map + MAP_NR(empty_zero_page))

/* number of bits that fit into a memory pointer */
#define BITS_PER_PTR	(8*sizeof(unsigned long))

/* to align the pointer to a pointer address */
#define PTR_MASK	(~(sizeof(void*)-1))

/* sizeof(void*) == 1<<SIZEOF_PTR_LOG2 */
#define SIZEOF_PTR_LOG2	2

#define page_address(page)  ({ if (!(page)->virtual) BUG(); (page)->virtual; }) 

/* SET_PAGE_DIR is used to set the segment table orign and length.
 * Although it loads cr1, note that it doesn't take effect until
 * DAT is set in the PSW. 
 */
/* tsk is a task_struct and pgdir is a pgd_t* */
#define SET_PAGE_DIR(tsk,pgdir)  { 					\
	(tsk)->thread.pg_tables = (unsigned long *)(pgdir);		\
	(tsk)->thread.cr1.raw = 0;					\
	(tsk)->thread.cr1.bits.psto = ((unsigned long) pgdir) >>12;	\
	(tsk)->thread.cr1.bits.pstl = 127;				\
	_lctl_r1 ((tsk)->thread.cr1.raw);				\
}
     
#define pte_ERROR(e) \
        printk("%s:%d: bad pte %08lx.\n", __FILE__, __LINE__, pte_val(e))
#define pmd_ERROR(e) \
        printk("%s:%d: bad pmd %08lx.\n", __FILE__, __LINE__, pmd_val(e))
#define pgd_ERROR(e) \
        printk("%s:%d: bad pgd %08lx.\n", __FILE__, __LINE__, pgd_val(e))


/* Unused pte's must have the invalid bit set ...  
 * Cleared PTE's must be marked invlaid, and have zeros in all other bit
 * positions.  This distinguishes them from swaped-out pte's, which don't 
 * contain 'true' ptes, but instead contain swapfile info.
 */

/*
 * Note this version was once hobbled by a compiler bug ...
 * extern inline int 
 * 	pte_none(pte_t pte)	{ return !((pte_val(pte) & ~(_PAGE_INVALID |_PAGE_RO))); }
*/
extern inline int 
	pte_none(pte_t pte)	{ return ((pte_val(pte) == (_PAGE_INVALID |_PAGE_RO))); }

extern inline int 
	pte_present(pte_t pte)	{ return !((pte_val(pte) & _PAGE_INVALID)); }
extern inline void 
	pte_clear(pte_t *ptep)	{ pte_val(*ptep) = _PAGE_INVALID | _PAGE_RO; }

extern inline unsigned long
	pte_pagenr(pte_t pte)	{ return (unsigned long) ((pte_val(pte) & PFRA_MASK) >> PAGE_SHIFT); }

/* unused segment table entries must have the invalid bit set.
 * XXX what's pmd_bad supposed to mean ????
 */
extern inline int 
	pmd_none(pmd_t pmd)	{ return pmd_val(pmd) & _SEG_INVALID; }
extern inline int 
	pmd_bad(pmd_t pmd)	{ return 0; }
extern inline int 
	pmd_present(pmd_t pmd)	{ return (pmd_val(pmd) & _SEG_INVALID) == 0; }
extern inline void 
	pmd_clear(pmd_t * pmdp)	{ pmd_val(*pmdp) = _SEG_INVALID; }


/*
 * The "pgd_xxx()" functions here are trivial for a folded two-level
 * setup: the pgd is never bad, and a pmd always exists (as it's folded
 * into the pgd entry)
 */
extern inline int pgd_none(pgd_t pgd)		{ return 0; }
extern inline int pgd_bad(pgd_t pgd)		{ return 0; }
extern inline int pgd_present(pgd_t pgd)	{ return 1; }
extern inline void pgd_clear(pgd_t * pgdp)	{ }

/*
 * The following only work if pte_present() is true.
 * Undefined behaviour if not..
 */
extern inline int pte_read(pte_t pte)		{ return pte_val(pte) & _PAGE_USER; }
extern inline int pte_write(pte_t pte)		{ return !(pte_val(pte) & _PAGE_RO); }
extern inline int pte_exec(pte_t pte)		{ return pte_val(pte) & _PAGE_USER; }

/* XXX we should get ref & change bits out of storage key */
extern inline int pte_dirty(pte_t pte)		{ return pte_val(pte) & _PAGE_DIRTY; }
extern inline int pte_young(pte_t pte)		{ return !(pte_val(pte) & _PAGE_ACCESSED); }

extern inline void pte_uncache(pte_t pte)       { }
extern inline void pte_cache(pte_t pte)         { }

extern inline pte_t 
	pte_rdprotect(pte_t pte) { pte_val(pte) &= ~_PAGE_USER; return pte; }
extern inline pte_t 
	pte_mkread(pte_t pte) 	{ pte_val(pte) |= _PAGE_USER; return pte; }

extern inline pte_t 
	pte_exprotect(pte_t pte) { pte_val(pte) &= ~_PAGE_USER; return pte; }
extern inline pte_t 
	pte_mkexec(pte_t pte) 	{ pte_val(pte) |= _PAGE_USER; return pte; }

extern inline pte_t 
	pte_wrprotect(pte_t pte) { pte_val(pte) |= _PAGE_RO ; return pte; }
extern inline pte_t 
	pte_mkwrite(pte_t pte) 	{ pte_val(pte) &= ~_PAGE_RO; return pte; }

extern inline pte_t 
	pte_mkclean(pte_t pte) 	{ pte_val(pte) &= ~(_PAGE_DIRTY ); return pte; }
extern inline pte_t 
	pte_mkdirty(pte_t pte) 	{ pte_val(pte) |= _PAGE_DIRTY; return pte; }

extern inline pte_t 
	pte_mkold(pte_t pte)  	{ pte_val(pte) |= _PAGE_ACCESSED; return pte; }
extern inline pte_t 
	pte_mkyoung(pte_t pte) 	{ pte_val(pte) &= ~_PAGE_ACCESSED; return pte; }



/* Certain architectures need to do special things when pte's
 * within a page table are directly modified.  Thus, the following
 * hook is made available. 
 * For the ESA/390, I think we're supposed to set/reset the
 * reference & change bits in the key storage at this time ...
 * XXX right ?? implement this ... 
 *
 * Note that the swapping code might call set_pte with an argument
 * of zero, in which case we need to set teh invalid bit.
 * (grep for set_pte (__pte(0));
 */
extern inline void 
set_pte (pte_t * ptep, pte_t val)	 
{
	if (pte_val(val)) { 
		*ptep = val;
	} else {
		pte_clear (ptep);
	}
}

/*
 * Conversion functions: convert a page and protection to a page entry,
 * and a page entry and page directory to the page they refer to.
 */

static inline pte_t mk_pte_phys(unsigned long page, pgprot_t pgprot)
{ pte_t pte; 
// printk ("mk_pte_phys for vaddr=%lx flags=%lx\n",page,pgprot_val(pgprot));
pte_val(pte) = (page & PFRA_MASK) | pgprot_val(pgprot); return pte; }

/*
 * mk_pte creates a pte entry from the pointer to a struct page.
 */
// extern inline pte_t mk_pte (struct page * page, pgprot_t pgprot)
#define mk_pte(page,pgprot)						\
({									\
	unsigned long pgnr = (unsigned long) ((page) - mem_map);	\
	unsigned long pg = pgnr << PAGE_SHIFT;				\
	pte_t pte; 							\
	/* printk ("mk_pte %lx %lx\n", pg, pgprot_val(pgprot));	*/	\
	pte_val(pte) = (pg & PFRA_MASK) | pgprot_val(pgprot); 		\
	pte; 								\
})

extern inline pte_t pte_modify(pte_t pte, pgprot_t newprot)
{ pte_val(pte) = (pte_val(pte) & _PAGE_CHG_MASK) | pgprot_val(newprot); 
// printk ("pte_modify %lx %lx\n", pte_val(pte),pgprot_val(newprot));
return pte; }

#define pte_page(x)             (mem_map+pte_pagenr(x))

/* pmd_page returns only the address part of the entry; mask out misc bits */
extern inline unsigned long pmd_page(pmd_t pmd)
{ return pmd_val(pmd) & SEG_PTO_MASK; }


/* to find an entry in a kernel segment-table-directory */
#define pgd_offset_k(address) pgd_offset(&init_mm, address)

/* to find an entry in a segment-table-directory */
extern inline pgd_t * pgd_offset(struct mm_struct * mm, unsigned long address)
{
	/* addend should be between 0 and 2048 */
	return mm->pgd + ((address & ADDR_MASK) >> PGDIR_SHIFT);
}

/* Find an entry in the second-level page table.. */
extern inline pmd_t * pmd_offset(pgd_t * dir, unsigned long address)
{
	return (pmd_t *) dir;
}

/* Find an entry in the third-level page table.. */ 
extern inline pte_t * pte_offset(pmd_t * dir, unsigned long address)
{
	pte_t * ppp = (pte_t *) pmd_page(*dir);
	/* addend should be between 0 and 256 */
	ppp += (address >> PAGE_SHIFT) & (PTRS_PER_PTE - 1);
	return ppp;
}

/* ===================================================== */

/* the pgdir consists of *TWO* pages (address 2GB of storage) */
extern pgd_t swapper_pg_dir[2048];

extern __inline__ pte_t *find_pte(struct mm_struct *mm, unsigned long va)
{
	pgd_t *dir;
	pmd_t *pmd;
	pte_t *pte = NULL;

	va &= ADDR_MASK;
	/* printk ("find_pte for va=0x%lx\n", va); */
	
	dir = pgd_offset( mm, va );
	if (dir)
	{
		/* printk ("find_pte got pgd for dir=%p\n", dir); */
		pmd = pmd_offset(dir, va);
		if (pmd && pmd_present(*pmd))
		{
			pte = pte_offset(pmd, va);
			/* printk ("find_pte got pmd for va=%lx pte=%p\n", va, pte); */
/* XXX the PowerPC implementation flushes the  tlb here, should we ?? */
		}
	}
	return pte;
}

/*
 * The ESA/390 architecture has no external MMU that needs to be updated
 */
#define update_mmu_cache(vma, addr, pte)	do { } while (0)


/* Page table entries that have been swapped out need to be marked invalid.
 * The other bits in the entry are used to store the swap location.
 * We'll have a 20-bit offset, which in practical terms means that
 * the max effective swapfile size is 2^20 pages = 4GBytes.
 * See for example try_to_swap_out() -> get_swap_page()
 *
 * Note that this scheme misidentifies swapfile 0, offset 0 as a 
 * non-existant pte (pte_none()==true).  
 */

#define SWP_TYPE(entry) 		((entry).val & 0xff)
#define SWP_OFFSET(entry) 		((entry).val >> 12)
#define SWP_ENTRY(type,offset) 		\
	((swp_entry_t) { (((offset) << 12) | _PAGE_INVALID | _PAGE_RO | (type)) } )
#define pte_to_swp_entry(pte)           ((swp_entry_t) { pte_val(pte) })
#define swp_entry_to_pte(x)             ((pte_t) { (x).val })

#define module_map      vmalloc
#define module_unmap    vfree

/* Needs to be defined here and not in linux/mm.h, as it is arch dependent */
#define PageSkip(page)		(0)

#endif __ASSEMBLY__
#endif /* _I370_PGTABLE_H */
